# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from designate import objects
from designate.objects.adapters.api_v2 import base
from designate import utils


class ZoneMasterAPIv2Adapter(base.APIv2Adapter):

    ADAPTER_OBJECT = objects.ZoneMaster

    MODIFICATIONS = {
        'fields': {
            'value': {
                'read_only': False
            }
        },
        'options': {
            'links': False,
            'resource_name': 'domain_master',
            'collection_name': 'domain_masters',
        }
    }

    @classmethod
    def _render_object(cls, object, *arg, **kwargs):
        if object.port == 53:
            return object.host
        else:
            return "%(host)s:%(port)d" % object.to_dict()

    @classmethod
    def _parse_object(cls, value, object, *args, **kwargs):
        object.host, object.port = utils.split_host_port(value)
        return object


class ZoneMasterListAPIv2Adapter(base.APIv2Adapter):

    ADAPTER_OBJECT = objects.ZoneMasterList

    MODIFICATIONS = {
        'options': {
            'links': False,
            'resource_name': 'domain_master',
            'collection_name': 'domain_masters',
        }
    }

    @classmethod
    def _render_list(cls, list_object, *args, **kwargs):

        r_list = []

        for object in list_object:
            r_list.append(cls.get_object_adapter(
                cls.ADAPTER_FORMAT,
                object).render(cls.ADAPTER_FORMAT, object, *args, **kwargs))

        return r_list

    @classmethod
    def _parse_list(cls, values, output_object, *args, **kwargs):

        for value in values:
            # Add the object to the list
            output_object.append(
                # Get the right Adapter
                cls.get_object_adapter(
                    cls.ADAPTER_FORMAT,
                    # This gets the internal type of the list, and parses it
                    # We need to do `get_object_adapter` as we need a new
                    # instance of the Adapter
                    output_object.LIST_ITEM_TYPE()).parse(
                        value, output_object.LIST_ITEM_TYPE()))

        # Return the filled list
        return output_object



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import dns
import dns.query
import dns.tsigkeyring
from oslo_config import cfg

from designate import dnsutils
from designate.mdns import handler
from designate import storage
import designate.tests

CONF = cfg.CONF


class TestSerializationMiddleware(designate.tests.TestCase):
    def setUp(self):
        super(TestSerializationMiddleware, self).setUp()
        storage_driver = CONF['service:central'].storage_driver
        self.storage = storage.get_storage(storage_driver)
        self.tg = mock.Mock()

    def test_with_tsigkeyring(self):
        self.create_tsigkey(fixture=1)

        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        query.use_tsig(dns.tsigkeyring.from_text(
            {'test-key-two': 'AnotherSecretKey'})
        )
        payload = query.to_wire()

        application = handler.RequestHandler(self.storage, self.tg)
        application = dnsutils.SerializationMiddleware(
            application, dnsutils.TsigKeyring(self.storage)
        )

        self.assertTrue(next(application(
            {'payload': payload, 'addr': ['192.0.2.1', 5353]}
        )))

    def test_without_tsigkeyring(self):
        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        payload = query.to_wire()

        application = handler.RequestHandler(self.storage, self.tg)
        application = dnsutils.SerializationMiddleware(
            application, dnsutils.TsigKeyring(self.storage)
        )

        self.assertTrue(next(application(
            {'payload': payload, 'addr': ['192.0.2.1', 5353]}
        )))


class TestTsigUtils(designate.tests.TestCase):
    def setUp(self):
        super(TestTsigUtils, self).setUp()
        storage_driver = CONF['service:central'].storage_driver
        self.storage = storage.get_storage(storage_driver)
        self.tsig_keyring = dnsutils.TsigKeyring(self.storage)

    def test_tsig_keyring(self):
        expected_result = b'J\x89\x9e:WRy\xca\xde\xb4\xa7\xb2'

        self.create_tsigkey(fixture=0)

        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        query.use_tsig(dns.tsigkeyring.from_text(
            {'test-key-one': 'SomeOldSecretKey'})
        )

        self.assertEqual(expected_result, self.tsig_keyring.get(query.keyname))
        self.assertEqual(expected_result, self.tsig_keyring[query.keyname])

    def test_tsig_keyring_not_found(self):
        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        query.use_tsig(dns.tsigkeyring.from_text(
            {'test-key-one': 'SomeOldSecretKey'})
        )

        self.assertIsNone(self.tsig_keyring.get(query.keyname))



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate import objects
from designate.tests import TestCase


LOG = logging.getLogger(__name__)


class ProducerServiceTest(TestCase):
    def setUp(self):
        super().setUp()
        self.producer_service = self.start_service('producer')

    def test_stop(self):
        self.producer_service.stop()

    def test_validate_partition_range(self):
        self.producer_service.start()

        min_partition = objects.Zone.fields['shard'].min
        max_partition = objects.Zone.fields['shard'].max

        self.assertIn(min_partition, self.producer_service.partition_range)
        self.assertIn(max_partition, self.producer_service.partition_range)



# Copyright 2016 Rackspace Inc.
#
# Author: Eric Larson <eric.larson@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.mport threading
import oslotest.base
from unittest import mock

from designate import exceptions
from designate import objects
from designate.worker.tasks import base


class TestTask(oslotest.base.BaseTestCase):
    def setUp(self):
        super(TestTask, self).setUp()
        self.context = mock.Mock()
        self.task = base.Task(None)
        self.storage = self.task._storage = mock.Mock()

    def test_constructor(self):
        self.assertTrue(self.task)

    def test_call(self):
        self.assertRaises(NotImplementedError, self.task)

    def test_current_action_is_valid(self):
        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='UPDATE')
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'UPDATE', objects.Zone(action='UPDATE'))
        )

        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='CREATE')
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'CREATE', objects.Zone(action='CREATE'))
        )

        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='UPDATE')
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'CREATE', objects.Zone(action='CREATE'))
        )

        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='DELETE')
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'DELETE', objects.Zone(action='DELETE'))
        )

    def test_current_action_delete_always_valid(self):
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'DELETE', None)
        )

    def test_current_action_bad_storage_always_valid(self):
        self.storage.get_zone = mock.Mock(
            side_effect=exceptions.DesignateException()
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'CREATE', objects.Zone(action='CREATE'))
        )

    def test_current_action_is_not_valid_none(self):
        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='NONE')
        )
        self.assertFalse(
            self.task.is_current_action_valid(
                self.context, 'UPDATE', objects.Zone(action='UPDATE'))
        )

    def test_current_action_is_not_valid_deleted(self):
        self.storage.get_zone = mock.Mock(
            return_value=objects.Zone(action='DELETE')
        )
        self.assertFalse(
            self.task.is_current_action_valid(
                self.context, 'UPDATE', objects.Zone(action='UPDATE'))
        )

    def test_current_action_is_not_found(self):
        self.storage.get_zone = mock.Mock(
            side_effect=exceptions.ZoneNotFound()
        )
        self.assertTrue(
            self.task.is_current_action_valid(
                self.context, 'CREATE', objects.Zone(action='CREATE'))
        )

        self.storage.get_zone = mock.Mock(
            side_effect=exceptions.ZoneNotFound()
        )
        self.assertFalse(
            self.task.is_current_action_valid(
                self.context, 'UPDATE', objects.Zone(action='UPDATE'))
        )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import socket
from unittest import mock

import dns
import dns.rdataclass
import dns.rdatatype

from designate import dnsutils
import designate.mdns.notify as notify
import designate.tests
from designate.tests.unit import RoObject


class MdnsNotifyTest(designate.tests.TestCase):
    def setUp(self):
        super(MdnsNotifyTest, self).setUp()

        self.notify = notify.NotifyEndpoint(mock.Mock())

    @mock.patch('time.sleep')
    def test_notify_zone_changed(self, mock_sleep):
        self.notify._make_and_send_dns_message = mock.Mock()

        self.notify.notify_zone_changed(*range(8))

        self.notify._make_and_send_dns_message.assert_called_with(
            1, 2, 3, 4, 5, 6, notify=True
        )

    @mock.patch('time.sleep')
    def test_get_serial_number_nxdomain(self, mock_sleep):
        # The zone is not found but it was supposed to be there
        response = RoObject(
            answer=[RoObject(
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.SOA
            )],
            rcode=mock.Mock(return_value=dns.rcode.NXDOMAIN)
        )
        zone = RoObject(name='zn', serial=314)
        self.notify._make_and_send_dns_message = mock.Mock(
            return_value=(response, 1)
        )

        out = self.notify.get_serial_number(
            'c', zone, 'h', 1234, 1, 2, 3, 4
        )

        self.assertEqual(('NO_ZONE', None, 0), out)

    @mock.patch('time.sleep')
    def test_get_serial_number_nxdomain_deleted_zone(self, mock_sleep):
        # The zone is not found and it's not was supposed be there
        response = RoObject(
            answer=[RoObject(
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.SOA
            )],
            rcode=mock.Mock(return_value=dns.rcode.NXDOMAIN)
        )
        zone = RoObject(name='zn', serial=0, action='DELETE')
        self.notify._make_and_send_dns_message = mock.Mock(
            return_value=(response, 1)
        )

        out = self.notify.get_serial_number(
            'c', zone, 'h', 1234, 1, 2, 3, 4
        )

        self.assertEqual(('NO_ZONE', 0, 3), out)

    @mock.patch('time.sleep')
    def test_get_serial_number_ok(self, mock_sleep):
        zone = RoObject(name='zn', serial=314)
        ds = RoObject(items=[zone])
        response = RoObject(
            answer=[RoObject(
                name='zn',
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.SOA,
                to_rdataset=mock.Mock(return_value=ds)
            )],
            rcode=mock.Mock(return_value=dns.rcode.NOERROR)
        )
        self.notify._make_and_send_dns_message = mock.Mock(
            return_value=(response, 1)
        )

        out = self.notify.get_serial_number(
            'c', zone, 'h', 1234, 1, 2, 3, 4
        )

        self.assertEqual(('SUCCESS', 314, 3), out)

    @mock.patch('time.sleep')
    def test_get_serial_number_too_many_retries(self, mock_sleep):
        zone = RoObject(name='zn', serial=314)
        ds = RoObject(items=[RoObject(serial=310)])
        response = RoObject(
            answer=[RoObject(
                name='zn',
                rdclass=dns.rdataclass.IN,
                rdtype=dns.rdatatype.SOA,
                to_rdataset=mock.Mock(return_value=ds)
            )],
            rcode=mock.Mock(return_value=dns.rcode.NOERROR)
        )
        self.notify._make_and_send_dns_message = mock.Mock(
            return_value=(response, 1)
        )

        out = self.notify.get_serial_number(
            'c', zone, 'h', 1234, 1, 2, 3, 4
        )

        self.assertEqual(('ERROR', 310, 0), out)

    @mock.patch('time.sleep')
    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_timeout(self, mock_send_dns_message,
                                               mock_sleep):
        zone = RoObject(name='zn')
        mock_send_dns_message.side_effect = dns.exception.Timeout

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((None, 3), out)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_bad_response(self,
                                                    mock_send_dns_message):
        zone = RoObject(name='zn')
        self.notify._make_dns_message = mock.Mock(return_value='')
        mock_send_dns_message.side_effect = notify.dns_query.BadResponse

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((None, 1), out)

    @mock.patch('time.sleep')
    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_eagain(self, mock_send_dns_message,
                                              mock_sleep):
        # bug #1558096
        zone = RoObject(name='zn')
        socket_error = socket.error()
        socket_error.errno = socket.errno.EAGAIN
        mock_send_dns_message.side_effect = socket_error

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((None, 3), out)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_econnrefused(self,
                                                    mock_send_dns_message):
        # bug #1558096
        zone = RoObject(name='zn')
        socket_error = socket.error()
        socket_error.errno = socket.errno.ECONNREFUSED
        # socket errors other than EAGAIN should raise
        mock_send_dns_message.side_effect = socket_error

        self.assertRaises(
            socket.error,
            self.notify._make_and_send_dns_message,
            zone, 'host', 123, 1, 2, 3
        )

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_nxdomain(self, mock_send_dns_message):
        zone = RoObject(name='zn')
        response = RoObject(rcode=mock.Mock(return_value=dns.rcode.NXDOMAIN))
        mock_send_dns_message.return_value = response

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((response, 1), out)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_missing_AA_flags(self,
                                                        mock_send_dns_message):
        zone = RoObject(name='zn')
        response = RoObject(
            rcode=mock.Mock(return_value=dns.rcode.NOERROR),
            # rcode is NOERROR but (flags & dns.flags.AA) gives 0
            flags=0,
            answer=['answer'],
        )
        mock_send_dns_message.return_value = response

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((None, 1), out)

    @mock.patch.object(dnsutils, 'send_dns_message')
    def test_make_and_send_dns_message_error_flags(self,
                                                   mock_send_dns_message):
        zone = RoObject(name='zn')
        response = RoObject(
            rcode=mock.Mock(return_value=dns.rcode.NOERROR),
            # rcode is NOERROR but flags are not NOERROR
            flags=123,
            ednsflags=321,
            answer=['answer'],
        )
        mock_send_dns_message.return_value = response

        out = self.notify._make_and_send_dns_message(
            zone, 'host', 123, 1, 2, 3
        )

        self.assertEqual((None, 1), out)

    def test_make_dns_message(self):
        msg = self.notify._make_dns_message('zone_name')
        txt = msg.to_text().split('\n')[1:]

        self.assertEqual([
            'opcode QUERY',
            'rcode NOERROR',
            'flags RD',
            ';QUESTION',
            'zone_name. IN SOA',
            ';ANSWER',
            ';AUTHORITY',
            ';ADDITIONAL'
        ], txt)

    def test_make_dns_message_notify(self):
        msg = self.notify._make_dns_message('zone_name', notify=True)
        txt = msg.to_text().split('\n')[1:]

        self.assertEqual([
            'opcode NOTIFY',
            'rcode NOERROR',
            'flags AA',
            ';QUESTION',
            'zone_name. IN SOA',
            ';ANSWER',
            ';AUTHORITY',
            ';ADDITIONAL',
        ], txt)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
from urllib import parse

from designate import exceptions
from designate.objects.adapters import base
from designate.objects import base as ovoobj_base


cfg.CONF.import_opt('api_base_uri', 'designate.api', group='service:api')
cfg.CONF.import_opt('enable_host_header', 'designate.api', group='service:api')


class APIv2Adapter(base.DesignateAdapter):

    BASE_URI = cfg.CONF['service:api'].api_base_uri.rstrip('/')

    ADAPTER_FORMAT = 'API_v2'

    #####################
    # Rendering methods #
    #####################

    @classmethod
    def _render_list(cls, list_object, *args, **kwargs):

        r_list = super(APIv2Adapter, cls)._render_list(
            list_object, *args, **kwargs)

        if cls.MODIFICATIONS['options'].get('links', True)\
                and 'request' in kwargs:
            r_list['links'] = cls._get_collection_links(
                list_object, kwargs['request'])
        # Check if we should include metadata
        if isinstance(list_object, ovoobj_base.PagedListObjectMixin):
            metadata = {}
            if list_object.total_count is not None:
                metadata['total_count'] = list_object.total_count
            r_list['metadata'] = metadata

        return r_list

    @classmethod
    def _render_object(cls, object, *args, **kwargs):
        obj = super(APIv2Adapter, cls)._render_object(object, *args, **kwargs)

        if cls.MODIFICATIONS['options'].get('links', True)\
                and 'request' in kwargs:
            obj['links'] = cls._get_resource_links(object, kwargs['request'])

        return obj

    #####################
    #  Parsing methods  #
    #####################

    @classmethod
    def parse(cls, values, output_object, *args, **kwargs):
        return super(APIv2Adapter, cls).parse(
            cls.ADAPTER_FORMAT, values, output_object, *args, **kwargs)

    #####################
    #    Link methods   #
    #####################

    @classmethod
    def _get_base_url(cls, request):
        if cfg.CONF['service:api'].enable_host_header:
            try:
                return request.host_url
            except Exception:
                return cls.BASE_URI
        else:
            return cls.BASE_URI

    @classmethod
    def _get_resource_links(cls, obj, request):
        base_uri = cls._get_base_url(request)

        path = cls._get_path(request, obj)
        return {'self': '%s%s/%s' % (base_uri, path, obj.id)}

    @classmethod
    def _get_path(cls, request, *args):
        path = request.path.lstrip('/').split('/')
        item_path = ''
        for part in path:
            if part == cls.MODIFICATIONS['options']['collection_name']:
                item_path += '/' + part
                return item_path
            else:
                item_path += '/' + part

    @classmethod
    def _get_collection_links(cls, item_list, request):

        links = {
            'self': cls._get_collection_href(request)
        }
        params = request.GET

        # defined in etc/designate/designate.conf.sample
        limit = cfg.CONF['service:api'].default_limit_v2

        if 'limit' in params:
            limit = params['limit']
            if limit.lower() == 'max':
                limit = cfg.CONF['service:api'].max_limit_v2
            else:
                try:
                    limit = int(limit)
                except ValueError:
                    raise exceptions.ValueError(
                        "'limit' should be an integer or 'max'")

        # Bug: this creates a link to "next" even on the last page if
        # len(item_list) happens to be == limit
        if limit is not None and limit == len(item_list):
            links['next'] = cls._get_next_href(request, item_list)

        return links

    @classmethod
    def _get_collection_href(cls, request, extra_params=None):
        params = request.GET

        if extra_params is not None:
            params.update(extra_params)

        base_uri = cls._get_base_url(request)

        href = "%s%s?%s" % (
            base_uri,
            cls._get_path(request),
            parse.urlencode(params))

        return href.rstrip('?')

    @classmethod
    def _get_next_href(cls, request, items):
        # Prepare the extra params
        extra_params = {
            'marker': items[-1]['id']
        }

        return cls._get_collection_href(request, extra_params)
